require 'enumerator'

class Array
  def delete_one(*x)
    if block_given?
      ndx = self.find_index{|item| yield(item)}
    else
      ndx = self.find_index(x.first)
    end
    self.delete_at(ndx) unless ndx.nil?
  end
  def include_array?(a)
    a.select{|item| !self.include?(item)}.empty?
  end
  # returns hash which has uniq values as keys and their multiplicity in value
  def multiplicity
    result = {}
    self.each do |item|
      result[item] = (result[item]||0) + 1
    end
    return result
  end
end

# takový temporary objekt, kdo kterého lze nacpat pole segmentů (tj pole polí koordinátů) a pak s tím šachovat a případně 
# si to výsledek ve stejném formátu převzít
class GpxWithGaps
  delegate :size, to: :@segments
  @segments = []  # array of GpxSegments
  @excursions = []

  def initialize(segment_coordinates)  # expects array of arrays of coordinates
    segment_coordinates = [] if segment_coordinates.nil?
    self.coordinates_array = segment_coordinates
  end

  def coordinates_array # getter
    @segments.collect { |segment| segment.coordinates_array }
  end
  
  def coordinates_array=(value) # setter
    if value.nil?
      coordinates_array = []
    else
      @segments = value.collect{ |coordinates_array| GpxSegment.new(coordinates_array) }
    end
    invalidate_cache
  end

  def invalidate_cache
    @coordinates_array = nil
  end
  
  # coordinaty vrcholu, pritom se nebere ohled na to co je gap a co neni
  def flat_coordinates_array
    @coordinates_array ||= coordinates_array
    @coordinates_array.flatten(1)
  end
  def flat; flat_coordinates_array end
  
  # sorts segments so the gaps are smallest possible + start coordinates sum is lower then end coordinates sum (this is here so the outcome is somehow normalized and nonrandom)
  # it also puts aside excursions
  def sort!
    endpoints = @segments.collect{|seg| seg.endpoints}.flatten(1)
    
    # create array of pairs of endpoints (not from the same segment)
    pairs = endpoints.combination(2).to_a
    @segments.each { |seg| pairs.delete(seg.endpoints) }
    
    # create hash containing distance of every pair of endpoints
    distance_hash = {}
    pairs.each {|pair| distance_hash[pair]=Map.distance(pair[0],pair[1]) }

    # sort pairs by distance
    pairs.sort!{|pairx,pairy| distance_hash[pairx]<=>distance_hash[pairy]} 

    ### go through sorted pairs and put aside the ones with not yet used endpoints
    unused_endpoints = Array.new(endpoints) # copy of array
    joints = []
    
    pairs.each do | pair |
      break if unused_endpoints.size<=2   # děláme tak dlouho dokud nezbydou dva a to nejvzdálenější endpointy

      if unused_endpoints.include_array?(pair)
        # ochrana proti cyklům  vždy přeuspořádáme pole unused_endpoints tak aby konce doposud spojených 
        # sekvencí byly v poli za sebou. A pak hlídáme aby se nikdy nespojily 
        i1 = unused_endpoints.find_index(pair.first).div(2)*2  # indexy zaokrouhlené na sudou
        i2 = unused_endpoints.find_index(pair.last).div(2)*2
        if i1 != i2                                             # nesmí jít o tutéž dvojici
          pom = unused_endpoints.slice!([i1,i2].max,2)          # vyhodí ven páry alignované na sudé indexy
          pom += unused_endpoints.slice!([i1,i2].min,2)         # vyhodí první vyšší index jinak by se pole posunulo 
          pom.delete_one(pair.first); pom.delete_one(pair.last) # z pom nechá jen nepoužité enpointy
          raise "jen 2" if pom.size!=2
          unused_endpoints += pom                               # a dá je na konec pole jako nový pár

          joints << GpxSegment.new(pair) 
        end
      end
    end
    
    @point_counter = 0
    # finally sort segments using joints + unused_endpoints (last 2 remaining)
    sorted_segments = []
    current_coordinates = unused_endpoints.first
    while true 
      # move segment to sorted_segments
      segment = @segments.delete_one { |seg| seg.has_endpoint?(current_coordinates) }
      
      break if segment.nil?
      segment.reverse! if segment.last == current_coordinates  # chceme aby segmenty byly správně otočeny
      sorted_segments << segment
      
      # find the joint and remove it from joints
      current_coordinates = segment.the_other_endpoint(current_coordinates)
      joint = joints.delete_one { |joint| joint.has_endpoint?(current_coordinates)}
      
      break if joint.nil? 
      current_coordinates = joint.the_other_endpoint(current_coordinates)
    end
    @segments = sorted_segments
    # order normalization
    set_normal_order
    
    invalidate_cache
    return self
  end
  
  def reverse!
    @segments = @segments.reverse.collect { |gpxsegment| gpxsegment.reverse! }
    invalidate_cache
  end

  def reversed?
    !@segments.first.nil? and !@segments.last.nil? and (@segments.first.first[0]+@segments.first.first[1] > @segments.last.last[0]+@segments.last.last[1])
  end
  
  def set_normal_order;  reverse! if reversed? end
  def set_reverse_order; reverse! if !reversed? end
  
  def gaps
    result = []
    @segments.each_cons(2){|a,b| result << [a.last,b.first]}
    result
  end

#   def nearest_point_to(point)
#     @segments.collect{|seg| nearest_point_to(point) }.min_by{ |candidate| Map.distance(candidate,point) }
#   end
  
  # index usecky, ke ktere ma bod nejblize. Cisla mezi 0 a 1 znamenaji, ze nejblize je k usecce 1 a desetinna mista znaci ke ktermu bodu na teto usecce. Napr 0.5 znamena prostredek.
  def line_parameter_of_nearest_point_to(apoint)
    parameter = 0
    min_distance = Float::INFINITY
    min_distance_parameter = 0
    
    self.flat_coordinates_array.each_cons(2) do |a,b|
      current_distance = Map.distance_of_point_from_segment(apoint,a,b)
      if current_distance < min_distance
        min_distance = current_distance
        fractionpart = Map.projection_parameter_of_point_to_segment(apoint,a,b)
        fractionpart = 0 if fractionpart<0 # když je nejblíž počátečnímu vrcholu tak nechceme mít záporný parametr
        fractionpart = 1 if fractionpart>1 # obdobně ke koncovému
        min_distance_parameter = parameter + fractionpart
      end
      parameter += 1
    end
    return min_distance_parameter
  end
  
  def kilometraz_of_nearest_point_to(apoint)
    param = line_parameter_of_nearest_point_to(apoint)
    result = (0..param.floor-1).sum { |i| length_of_line_segment(i) }
    fractionpart  = 
      if (param.modulo(1) == 0)
        then 0  # jinak by to počítalo zbytečně 0 krát error
        else param.modulo(1)*length_of_line_segment(param.floor)  
      end
    return (result + fractionpart)/1000 # v kilometrech
  end

  def nearest_point_to(apoint)
    min_distance = Float::INFINITY
    nearest_point = nil
    
    self.flat_coordinates_array.each_cons(2) do |a,b|
      current_point = Map.nearest_point_on_segment(apoint,a,b)
      current_distance = Map.distance(apoint,current_point)
      if current_distance < min_distance
        nearest_point = current_point
        min_distance = current_distance
      end
    end
    return nearest_point
  end
  def distance_to(apoint); Map.distance(apoint,nearest_point_to(apoint)) end
  
  # 0 based
  def length_of_line_segment(index)
    fca = flat_coordinates_array
    Map.distance(fca[index],fca[index+1]) 
  end
  
  # vrati bod na trail se zadanou kilometrazi
  def point_with_kilometraz(kms)
    kms = 0 if kms<0
    metry = kms*1000
    distance = 0
    (0..flat_coordinates_array.size-2).each do |index|
      segment_length = length_of_line_segment(index)
      if distance + segment_length > metry
        line_segment_parameter = (metry-distance)/segment_length
        return Map.point_on_line_segment_with_parameter(flat_coordinates_array[index],flat_coordinates_array[index+1],line_segment_parameter)
      else
        distance += segment_length
      end
    end
    return flat_coordinates_array.last
  end
  
  def length_with_gaps # in metres
    sum = 0
    flat_coordinates_array.each_cons(2) do |a,b|
      sum += Map.distance(a,b)
    end
    sum
  end
  
  def length_without_gaps # in metres
    @segments.sum{ |segment| segment.length }
  end
  
  def first_point; @segments.andand.first.andand.first end
  def last_point; @segments.andand.last.andand.last end
  
  def endpoints
    @segments.collect_concat{ |segment| segment.endpoints }
  end

  def center_point # guess apporximate center of gpx
    # average of endpint coordinates
    e = endpoints
    e.reduce{ |sum,it| [sum[0]+it[0], sum[1]+it[1]] }.map{ |r| r/e.size }
  end
  
  # pole bodů, kde se gpx větví - tj, body, které sousedí s více než dvěma body
  def branching_points
    multiplicity_hash = flat_coordinates_array.multiplicity

    pole_endpointu = endpoints
    
#     if pole_endpointu.uniq.count != pole_endpointu.count
#       raise "Pole endpointu by nemelo obsahovat jednu polozku dvakrat, ale v tomto pripade je pole endpointu #{pole_endpointu.inspect}"
#     end
    
    # branching jsou body co jsou ve flat_coordinates_array aspoň dvakrát
    # i kdyby tenhle bod byl endpoint, tak to bude maximalne jednou, takze ten druhy vyskyt musi byt vprostred segmentu
    return multiplicity_hash.select{ |k,v| v > 1 }.keys

  end
end





















